/*
** aqua.h for include in include
**
** Made by William Caldwell aka
** Login   <William@undefined.re>
**
** Started on  Sun 30 Nov 2014 11:55:43 AM CET William Caldwell
** Last update Sun 30 Nov 2014 11:55:43 AM CET William Caldwell
*/

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#define in_port_t u_short
#define ssize_t int
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#endif

#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#ifndef	AQUA_H_
# define AQUA_H_

# define _unused __attribute__ ((unused))

// from aqua to here: reference node by index in array. (check i >=0 && < MAX_NODES)
# define MAX_NODES				210
# define MAX_FP_LOCAL_CIRCS			210
# define MAX_FP_MIX_CIRCS			200
# define MAX_QUEUE				10
# define MAX_NB_SECRETS				110 // max nb hops for client. this should be 1 for mixes.
# define MAX_CHANS				10 // max nb hops for client. this should be 1 for mixes.
# define AQ_BUFFER_MAX				512
# define AQUA_PACKET_SIZE			400
# define AQUA_SECRET_SIZE			32
# define AQUA_IV_SIZE				16
# define AQUA_CHAFFING_TIME			10000
//# define AQUA_CHAFFING_TIME			1000000

# define MAX_CH_CLIENTS				10

# define AQUA_SP_CLIENT_PACKET_SIZE		AQUA_PACKET_SIZE + sizeof (struct sp_client_header)
# define AQUA_SP_MANIFEST_PACKET_SIZE           AQUA_PACKET_SIZE + sizeof (struct sp_manifest_header)

//# define AQUA_SP_CLIENT_PACKET_SIZE		(AQUA_PACKET_SIZE + 24)
#// define AQUA_SP_MANIFEST_PACKET_SIZE		(AQUA_PACKET_SIZE + 24 * MAX_CH_CLIENTS)

// for openssl EVP:
# define ENCRYPT				1
# define DECRYPT				0

// FIXME fflush only for debug
# define LOGC(format, ...) do {						\
		fprintf(log_file, format "\n	-- %s:	%s:%d\n",	\
			##__VA_ARGS__ ,  __func__, __FILE__, __LINE__);	\
		fflush(log_file);					\
	} while (0)

# define LOG(format, ...) do {} while (0)
# define LOGP LOG

// cross compile

// mutexes:

# ifdef WIN32
#  define	MUTEX		HANDLE
#  define	THREAD_RED	DWORD WINAPI
#  define	THREAD_ARG	LPVOID
# else 
#  define	MUTEX		pthread_mutex_t
#  define	THREAD_RED	void
#  define	THREAD_ARG	void
# endif

int MUTEX_INIT(MUTEX *mutex);
int MUTEX_LOCK(MUTEX *mutex);
int MUTEX_UNLOCK(MUTEX *mutex);
void MUTEX_DESTROY(MUTEX *mutex);


// roles:
enum
{
	AQUA_APP_PROXY = 0,
	AQUA_MIX,
	AQUA_SIP_DIR,
	AQUA_RDV,
	AQUA_SP,
	AQUA_APP_PROXY_DUMMY,
	MAX_ROLES,
};

enum
{
	AQUA_CMD_INIT = 0,
	AQUA_CMD_CONNECT_TO_NODE,
	AQUA_CMD_OPEN_LOCAL_UDP,
	AQUA_CMD_UPDATE_LOCAL_UDP_DEST,
	AQUA_CMD_FORWARD,
	AQUA_CMD_DATA,
	AQUA_CMD_NEW_CIRCUIT,
	AQUA_CMD_ACK,
	AQUA_CMD_NEW_CLIENT,
	AQUA_CMD_RM_NODE,
	AQUA_CMD_NEW_MIX_FP,
	AQUA_CMD_RM_MIX_FP,
	AQUA_CMD_RM_LOCAL_UDP,
	AQUA_CMD_UPDATE_ROLE,
	AQUA_CMD_UPDATE_NODE_SECRET,
	AQUA_CMD_PING,
};
#define AQUA_CMD_REGISTER_ID_TO_SP	3

typedef					union
{
	struct sockaddr_storage		ss;
	struct sockaddr_in6		s6;
	struct sockaddr_in		s4;
}					addr;

struct					circ
{
	int				id;
	int				index; // for local fastpath: when in, index = local sock, when out, index of dtls sock. FIXME: no, if set it's always a dtls sock index.
	int				nb_secrets;
	uint8_t *			secrets[MAX_NB_SECRETS];
};

struct					local_fp
{
	int				used;
	int				fd;
	struct circ			circ_in;
	struct circ			circ_out;
	addr				local_in_peer;
	uint16_t			port;
	pthread_t			write_tid;
};

struct					mix_fp
{
	int				used;
	struct circ			circ_fwd;
	struct circ			circ_bwd;
};

struct					channel
{
	uint8_t *			client_secrets[MAX_CH_CLIENTS];
	int				client_node_index[MAX_CH_CLIENTS];
	int				used[MAX_CH_CLIENTS];
};

struct				sp_client_header
{
	uint8_t			iv[AQUA_IV_SIZE];
	uint32_t		seq;
	uint32_t		msg_type;
} __attribute__ ((packed));

struct				sp_manifest_header
{
	uint8_t			chan_id;
	struct sp_client_header	client[MAX_CH_CLIENTS];
} __attribute__ ((packed));

struct					node_info
{
	int				used;
	int				index;
	int				role;
	addr				peer_addr;
	addr				our_addr;
	pthread_t			write_tid;
	pthread_t			read_tid;
	pthread_mutex_t			ssl_lock;
	SSL *				ssl;
	// outgoing
	pthread_mutex_t			queue_lock;
	uint8_t *			queue[MAX_QUEUE];
	int				queue_len[MAX_QUEUE];
	int				queue_head;
	int				queue_tail;
	// incoming
	pthread_mutex_t			in_queue_lock;
	uint8_t				in_queue[AQUA_PACKET_SIZE + AQUA_SP_MANIFEST_PACKET_SIZE];
	int				in_q_sz;
	// other
	int				cookie;
	int				fd;
	// for superpeers:
	uint8_t *			secret;
	int				client_index;
 	// for SP & mix
	struct channel			channels[MAX_CHANS];
	// for clients:
	uint32_t			chan_id;

	// tmp, just ideas of what we might need:
	pthread_mutex_t			aqua_lock; // for editing circuits and such
	uint8_t *			ntor_id;
	void *				open_circuits;
};

struct				sp_client
{
	int			local_index;
	int			id;
};

struct				state
{
	uint8_t *		cert_file;
	uint8_t *		key_file;
	int			ntor_sec_size;
	int			ntor_pub_size;
	int			ntor_id_size;
	uint8_t *		ntor_id;
	uint8_t *		ntor_sec;
	uint8_t *		ntor_pub;
	int			aqua_dtls_port;
	int			aqua_socket;
	int			role;
	addr			aqua_peer;
	int			aqua_peer_len;
	pthread_t		server_tid;
	const EVP_CIPHER *	cipher;
	// sp only:
	//struct channel		channel[MAX_CHANS];
	// things. might malloc these during init to make maxes configurable.
	struct node_info	nodes[MAX_NODES];
	struct local_fp		local_fp[MAX_FP_LOCAL_CIRCS];
	struct mix_fp		mix_fp[MAX_FP_MIX_CIRCS];
};


uint8_t *
mix_unxor(uint8_t *		message,
	  int			len,
	  struct channel *	channel);

int
xor_to_queue(struct node_info *		node,
	     int			client_index,
	     uint8_t *			client_msg);

int
send_cell_to_sp(uint8_t *	message,
		uint8_t *	secret,
		int		msg_type,
		int		node_index);

int
get_free_node(struct state *		state);

int
create_rated_write_thread(struct node_info *		node);

void
alloc_and_copy_string(const uint8_t *	source,
		      uint8_t **	dest,
		      int		len);

void
print_state();

int
send_ack(uint32_t	cookie,
	 uint32_t	ret_value,
	 uint32_t	id);

int
send_data_packet(uint8_t *		msg,
		 int			len,
		 struct node_info *	node);

int
send_new_client(struct node_info *	node);

int
send_rm_node(struct node_info *	node);

int
process_fastpath(uint8_t *		message,
		 uint32_t		len,
		 struct node_info *	node);

int
process_init(uint8_t *		message,
	     int		len);

uint8_t *
parse_addr(uint8_t *		message,	// destructive on message.
	   int			len,		// FIXME: not using len because we know we set a 0 at the end of the buffer at the recvfrom. might change this.
	   addr *		dest);

int
process_connect_to_node(uint8_t *	message,
			int		len);

int
process_aqua_message(uint8_t *	message,
		     int	len,
		     int	aqua_socket,
		     addr *	aqua_peer,
		     socklen_t	aqua_peer_len);

void
dummy_mix_add_sp(struct node_info * node);


const char *
role_to_string(int	role_index);

#endif /* !AQUA_H_ */
